/*
 *
 * Copyright (c) 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Hourglass.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation is hereby granted without fee, provided that the
 * above copyright notice and this permission/disclaimer notice is
 * retained in all copies or modified versions, and that both notices
 * appear in supporting documentation.  THE COPYRIGHT HOLDERS PROVIDE
 * THIS SOFTWARE "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE COPYRIGHT
 * HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu
 * any improvements that they make and grant redistribution rights to the
 * University of Utah.
 * 
 * Author: John Regehr (regehr@cs.utah.edu)
 */

#include "hourglass.h"

#ifdef GAP_HIST
static void record_gap (cycle_time gap) 
{
  if (gap/GAP_HIST_DIVISOR > NUM_GAP_HIST_BUCKETS) {
    printf ("hrm: gap %lld exceeds NUM_GAP_HIST_BUCKETS %d\n",
	    gap/GAP_HIST_DIVISOR, NUM_GAP_HIST_BUCKETS);
    exit (-1);
  }
  gap_hist[gap/GAP_HIST_DIVISOR]++;
}
#endif

static int get_new_rec (void)
{
  long n;

  n = InterlockedIncrement (&current_record);

  if (n >= NUM_TRACE_RECORDS) {
    printf ("oops: trace record overrun - exiting\n");
    exit (-1);
  }

  return n;
}

static cycle_time poll (int id, int *pos)
{
  cycle_time now, gap, runtime;

  thrd_data[id].workdone++;

  if (pos) {
    thrd_data[id].scanmem[*pos] = (-2*thrd_data[id].scanmem[*pos+1]) + 77;
    (*pos)++;
    if ((*pos)+1 == (thrd_data[id].memsize / sizeof (int))) *pos = 0;
  }

  now = rdtsc ();
  gap = now - thrd_data[id].last_exec;

  if (gap > GAP) {
    if (thrd_data[id].exec_start == 0) {
      thrd_data[id].reserved_record = get_new_rec ();
    } else {
      int rec = thrd_data[id].reserved_record;
      thrd_data[id].reserved_record = get_new_rec ();
      trace_recs[rec].id = id;
      trace_recs[rec].start = 
	thrd_data[id].exec_start + 
	us_to_cycle (POLL_CORRECTION);
      trace_recs[rec].end = 
	thrd_data[id].last_exec - 
	us_to_cycle (POLL_CORRECTION);
    }
    thrd_data[id].exec_start = now;
    runtime = us_to_cycle(2*POLL_CORRECTION);
  } else {
    runtime = gap;
#ifdef GAP_HIST
    record_gap (gap);
#endif
  }

  thrd_data[id].last_exec = now;
  
  return runtime;
}

static void run_for (int id, cycle_time amount, int *pos)
{
  cycle_time total = 0;

  do {
    total += poll (id, pos);
  } while (total < amount && thrd_data[id].status != STOP);
}

#define FUDGE 10

static void periodic_test (int id)
{
  cycle_time now, deadline;

  deadline = rdtsc();

  while (thrd_data[id].status != STOP) {

    if (thrd_data[id].work == LAT) {
      if (thrd_data[id].lat_slot < LAT_SLOTS) {
	thrd_data[id].lat_slots[thrd_data[id].lat_slot] = rdtsc() - deadline;
	thrd_data[id].lat_slot++;
      }
    } else {
      run_for (id, us_to_cycle (thrd_data[id].app_amount), NULL);
    }

    deadline += us_to_cycle (thrd_data[id].app_period);      
    now = rdtsc();

    if (now <= deadline) {
      thrd_data[id].hits++;
      thrd_data[id].timers->timed_sleep (id, mymin (0, cycle_to_us (now - deadline) + FUDGE));
    } else {
      thrd_data[id].misses++;
      deadline = now;
    }
    
  }
}

static void cpu_yield_test (int id, int scan)
{
  int pos = 0;

  while (thrd_data[id].status != STOP) {
    run_for (id, us_to_cycle (thrd_data[id].app_amount), (scan) ? &pos : NULL);
    sched_yield ();
  }
}

static void cpu_test (int id, int scan)
{
  int pos = 0;

  while (thrd_data[id].status != STOP) {
    poll (id, (scan) ? &pos : NULL);
  }
}

void *monitor_thrd_func (void *arg)
{
  int id = (int)arg;

  assert (thrd_data[id].s_time != -1);

  usleep ((unsigned int)(thrd_data[id].f_time));  

  if (thrd_data[id].status == GO) {
    thrd_data[id].status = STOP;
  }

  /*
   * now, sleep a long time -- do this instead of just exiting since
   * exiting might be slow -- sleeping monitor threads will get
   * cleaned up when main() exits 
   */
  sleep (1000000000);

  return NULL;
}

void *thrd_func (void *arg)
{
  int id = (int)arg;

  // printf ("mypid is %d\n", getpid());

#if (defined WITH_CMU_RK || defined WITH_TIMESYS_RK)
  if (thrd_data[id].res_period != -1) {
    make_res (thrd_data[id].res_amount, thrd_data[id].res_period, id);
  }
#endif

  set_priority (thrd_data[id].pri);

  if (thrd_data[id].s_time != -1) {
    thrd_data[id].timers->timed_sleep (id, -thrd_data[id].s_time);  
  }

  thrd_data[id].workdone = 0;

  switch (thrd_data[id].work) {

  case LAT:
  case PERIODIC:
  case CPU_PERIODIC:
    thrd_data[id].hits = thrd_data[id].misses = 0;
    periodic_test (id);
    break;
    
  case CPU:
    cpu_test (id, FALSE);
    break;
    
  case CPU_SCAN:
    if (thrd_data[id].memsize > 0) {
      cpu_test (id, TRUE);
    } else {
      cpu_test (id, FALSE);
    }
    break;

  case CPU_YIELD:
    cpu_yield_test (id, FALSE);
    break;
    
  case CPU_SCAN_YIELD:
    if (thrd_data[id].memsize > 0) {
      cpu_yield_test (id, TRUE);
    } else {
      cpu_yield_test (id, FALSE);
    }
    break;

  default:
    printf ("Hrm: thread %d has an unknown workload\n", id);
    exit (-1);
  }

  assert (thrd_data[id].status == STOP);

  thrd_data[id].status = DONE;

#if (defined WITH_CMU_RK || defined WITH_TIMESYS_RK)
  delete_res (id);
#endif

  return NULL;
}
